/*
  default pin assign 

  um232r.cbl_type  = ft245r
  um232r.tms   = 7   # D7/RI
  um232r.tck   = 5   # D5/DSR
  um232r.tdi   = 6   # D6/DCD
  um232r.tdo   = 3   # D3/CTS

  um232r.init     = 11  # CB3
  um232r.prog     = 10  # CB2
  um232r.m2      = 9   # CB1
  um232r.m1      = 8   # CB0

  mpsse.cbl_type  = ft245r
  mpsse.tck   = 0   # ADBUS0/TXD
  mpsse.tdi   = 1   # ADBUS1/RXD
  mpsse.tdo   = 2   # ADBUS2/~RTS
  mpsse.tms   = 3   # ADBUS3/~CTS
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#include "util.h"
#include "cable.h"
#include "ft245r.h"

//#define USE_LIBFTDI

#if defined(_WIN32) || defined(SUPPORT_FT245R)
#if defined(_WIN32)
#include <windows.h>
#include "ftd2xx.h"
#elif defined(USE_LIBFTDI)
#include "ftdi.h"
#define FT_STATUS	int
#define DWORD 		uint32_t
#define UCHAR		uint8_t
#define FT_OK		(0)
#define FT_ERROR	(-1)
#else
#include "ftd2xx.h"
#endif

#define FT245R_CYCLES		 2
#define FT245R_FRAGMENT_SIZE  (is_hispeed?(62*64*4):(62*4))
#define FT245R_MAX_FRAGMENT_SIZE  (62*64*4)
#define REQ_OUTSTANDINGS	    10

#define FT245R_DEBUG	1

extern char * progname;
extern int do_cycles;
extern int ovsigck;
extern int verbose;

static int req_count = 0;

#if defined(USE_LIBFTDI)
static struct ftdi_context *handle;
#else
static FT_HANDLE handle;
#endif

static uint8_t dbus_ddr = 0;
static uint8_t dbus_data = 0;

static uint8_t dbus_tck = 0;
static uint8_t dbus_tdi = 0;
static uint8_t dbus_tms = 0;
static uint8_t dbus_tdo = 0;

static uint8_t dbus_prog  = 0;
static uint8_t dbus_init  = 0;
static uint8_t dbus_m1 = 0;
static uint8_t dbus_m2 = 0;

static uint8_t prog_is_neg = 0;
static uint8_t init_is_neg = 0;
static uint8_t m1_is_neg = 0;
static uint8_t m2_is_neg = 0;

static uint8_t cbus_ddr = 0;
static uint8_t cbus_data = 0;

static uint8_t cbus_prog = 0;
static uint8_t cbus_init = 0;
static uint8_t cbus_m1 = 0;
static uint8_t cbus_m2 = 0;

static int is_hispeed = 0;
extern char *conf_getenv(char *);

static void setup_pin_conf(CABLE cbl, char *name, uint8_t *data, 
			uint8_t *cbus_data, uint8_t *is_neg) {
	char *v;
	int pin;
	char full_name[256];

	strcpy(full_name,cbl->cbl_name);
	strcat(full_name,".");
	strcat(full_name,name);

	if (data) *data = 0;
	if (cbus_data) *cbus_data = 0;
	if (is_neg) *is_neg = 0;


	v = conf_getenv(full_name);
	if (!v) return;
	if (*v == '~') {
		if (is_neg) *is_neg = 1;
		v++;
		while (*v == ' ') v++;
	}
	pin = atoi(v);
	if ((0 <= pin) && (pin < 8)) *data = 1 << pin;
	else if ((8 <= pin) && (pin < 12)) *cbus_data = 1 << (pin - 8);

	return;
}

static int ft245r_send(CABLE cbl, uint8_t * buf, size_t len)
{
  FT_STATUS r = FT_OK;
  DWORD rlen;
#if !defined(USE_LIBFTDI)
  r = FT_Write(handle, buf, len, &rlen);
  if (r == FT_OK) return 0;
  if (len != rlen) return -1;
#else
  r = rlen = ftdi_write_data(handle, buf, len);
  if (r >= 0) r = FT_OK;
  if (len != rlen) r = FT_ERROR;
#endif
  return r;
}

static int ft245r_recv(CABLE cbl, uint8_t * buf, size_t len)
{
  FT_STATUS r = FT_OK;
  DWORD rlen;

#if !defined(USE_LIBFTDI)
  r = FT_Read(handle, buf, len, &rlen);
#else
retry:
  r = rlen = ftdi_read_data(handle, buf, len);
  if (r >= 0) {
	r = FT_OK;
  	if (len != rlen) {
		buf += rlen;
		len -= rlen;
		goto retry;
	}
  }
#endif

  if (r != FT_OK || len != rlen) {
    fprintf(stderr,
	    "%s: recv(): programmer is not responding\n",
	    progname);
    exit(1);
  }
  return 0;
}

static int drain(CABLE cbl, int display)
{
  FT_STATUS r = FT_OK;
#if !defined(USE_LIBFTDI)
  DWORD n;
  r = FT_SetBitMode(handle, 0, FT_BITMODE_RESET);
  if (r != FT_OK) return -1;
  r = FT_GetQueueStatus(handle, &n);
  if (r != FT_OK) return -1;
  if (n) {
	fprintf(stderr, "drain called but queue is not empty %d \n",
		(int)n);
  }
  r = FT_SetBitMode(handle, 0, FT_BITMODE_SYNC_BITBANG);
  if (r != FT_OK) return -1;
  r = FT_SetBaudRate(handle, 300); // 600 Hz : period H = 0.833 ms
  if (r != FT_OK) return -1;

  r = FT_SetBitMode(handle, dbus_ddr, FT_BITMODE_SYNC_BITBANG);
  if (r != FT_OK) return -1;

#else
  r = ftdi_set_bitmode(handle, 0, BITMODE_RESET);
  if (r != FT_OK) return -1;
  r = ftdi_usb_purge_buffers(handle);
  if (r != FT_OK) return -1;
  r = ftdi_set_bitmode(handle, 0, BITMODE_SYNCBB);
  if (r != FT_OK) return -1;
  r = ftdi_set_baudrate(handle, 1200); // 2400 Hz : period 0.208 ms
  if (r != FT_OK) return -1;

  r = ftdi_set_bitmode(handle, dbus_ddr, BITMODE_SYNCBB);
  if (r != FT_OK) return -1;
#endif
  return 0;
}

static inline int sync(CABLE cbl)
{
  FT_STATUS r = FT_OK;
#if !defined(USE_LIBFTDI)
  UCHAR ch;
  r = FT_GetBitMode(handle, &ch);
#else
  ftdi_async_complete(handle , 1);
#endif
  if (r != FT_OK) return -1;
  return 0;
}

static int valid_rates_r[] = {	// FT245R,  FT232R , FT2232D 
   600, 1200, 2400, 4800, 9600, 14400, 19200, 38400, 57600, 
   115200, 230400, 460800, 921600
//   , 1000000, 1500000, 3000000
};

static int valid_rates_h[] = {	// FT2232H , FT232H
   600, 1200, 2400, 4800, 9600, 14400, 19200, 38400, 57600, 
   115200, 230400, 460800, 921600,
   1000000, 1200000, 1500000, 2000000, 2400000, 3000000
//   4000000, 6000000, 12000000
};

static int *valid_rates = valid_rates_r;
int valid_rates_num = sizeof(valid_rates_r)/sizeof(valid_rates_r[0]);

static int saved_rate;

static int set_bitclock(CABLE cbl, double bitclock) {
  FT_STATUS r = 0;
  int rate = 0;
  int i;

  if (bitclock == 0.0) { // using default
	  rate = 235000.0 *3/8; 
  } else if (bitclock <  1500.0) {
	  rate =   1500.0 *3/8;
  } else {
	  rate = bitclock * 3/8 + 1.0;
  }
  for (i= valid_rates_num -1; i>=0; --i) 
  {
    if (valid_rates[i] <= rate) {
		rate = valid_rates[i];
		break;
    }
  }
  if (i<=0) rate = valid_rates[0];

#if !defined(USE_LIBFTDI)
  r = FT_SetBaudRate(handle, rate);
#else
  r = ftdi_set_baudrate(handle, rate);
#endif
  if ((verbose>=1) || FT245R_DEBUG) {
	fprintf(stderr," ft245r:  bitclk %d -> ft baud %d\n", 
			rate * 8/3, rate);
  }
  cbl->bitclock = rate * 8/3;
  cbl->baudrate = rate;
  saved_rate = rate;
  return 0;
}

#if 0
static int restore_bitclock(CABLE cbl) {
  FT_STATUS r;
#if !defined(USE_LIBFTDI)
  r = FT_SetBaudRate(handle, saved_rate);
#else
  r = ftdi_set_baudrate(handle, saved_rate);
#endif
  return r;
}
#endif

#define DELAY_LEN	(is_hispeed?(62*64*4):(62*8))
#define MAX_DELAY_LEN	(62*64*4)
static void ft245r_delay(CABLE cbl, int us)
{
  int len, sub_len;
  int send_count = 0;
  int recv_count = 0;
  uint8_t buf[MAX_DELAY_LEN];
  uint8_t dmy[MAX_DELAY_LEN];

  memset(buf, dbus_data, DELAY_LEN);

  len = ( (double)us * saved_rate * 16/3 + 999999.0 ) / 1000000.0;
  if (len < 1) len = 1;
  while (len > 0) {
    sub_len = len;
    if (sub_len > DELAY_LEN)
	sub_len = DELAY_LEN;
    len -= sub_len;

    ft245r_send (cbl, buf, sub_len);
    send_count++;
    if (send_count > REQ_OUTSTANDINGS) {
	ft245r_recv (cbl, dmy, DELAY_LEN);
        recv_count++;
    }
  }
  while (send_count > recv_count) {
        recv_count++;
	ft245r_recv (cbl, dmy, (send_count==recv_count)?sub_len:DELAY_LEN);
  }
}

static int setup_gpio(CABLE cbl, int bit, int value)
{
  FT_STATUS r = 0;
  uint8_t wk;
  int data_changed = 0;
  int cbus_data_changed = 0;

 if (bit == CABLE_GPIO_PROG) {
	if (dbus_prog) {
		if (value < 0) {
			dbus_ddr  &= ~dbus_prog;
			dbus_data &= ~dbus_prog;
		} else if ((value && !prog_is_neg)
		   || (!value && prog_is_neg)) {
			dbus_ddr  |= dbus_prog;
			dbus_data |= dbus_prog;
		} else {
			dbus_ddr  |= dbus_prog;
			dbus_data &= ~dbus_prog;
		}
  		data_changed = 1;
	}
	if (cbus_prog) {
		if (value < 0) {
			cbus_ddr  &= ~cbus_prog;
			cbus_data &= ~cbus_prog;
		} else if ((value && !prog_is_neg)
		   || (!value && prog_is_neg)) {
			cbus_ddr  |= cbus_prog;
			cbus_data |= cbus_prog;
		} else {
			cbus_ddr  |= cbus_prog;
			cbus_data &= ~cbus_prog;
		}
  		cbus_data_changed = 1;
	}
 } else if (bit == CABLE_GPIO_M1) {
	if (dbus_m1) {
		if (value < 0) {
			dbus_ddr  &= ~dbus_m1;
			dbus_data &= ~dbus_m1;
		} else if ((value && !m1_is_neg)
		   || (!value && m1_is_neg)) {
			dbus_ddr  |= dbus_m1;
			dbus_data |= dbus_m1;
		} else {
			dbus_ddr  |= dbus_m1;
			dbus_data &= ~dbus_m1;
		}
  		data_changed = 1;
	}
	if (cbus_m1) {
		if (value < 0) {
			cbus_ddr  &= ~cbus_m1;
			cbus_data &= ~cbus_m1;
		} else if ((value && !m1_is_neg)
		   || (!value && m1_is_neg)) {
			cbus_ddr  |= cbus_m1;
			cbus_data |= cbus_m1;
		} else {
			cbus_ddr  |= cbus_m1;
			cbus_data &= ~cbus_m1;
		}
  		cbus_data_changed = 1;
	}
 } else if (bit == CABLE_GPIO_M2) {
	if (dbus_m2) {
		if (value < 0) {
			dbus_ddr  &= ~dbus_m2;
			dbus_data &= ~dbus_m2;
		} else if ((value && !m2_is_neg)
		   || (!value && m2_is_neg)) {
			dbus_ddr  |= dbus_m2;
			dbus_data |= dbus_m2;
		} else {
			dbus_ddr  |= dbus_m2;
			dbus_data &= ~dbus_m2;
		}
  		data_changed = 1;
	}
	if (cbus_m2) {
		if (value < 0) {
			cbus_ddr  &= ~cbus_m2;
			cbus_data &= ~cbus_m2;
		} else if ((value && !m2_is_neg)
		   || (!value && m2_is_neg)) {
			cbus_ddr  |= cbus_m2;
			cbus_data |= cbus_m2;
		} else {
			cbus_ddr  |= cbus_m2;
			cbus_data &= ~cbus_m2;
		}
  		cbus_data_changed = 1;
	}
 }

  if (cbus_data_changed) {
     wk = (cbus_ddr << 4) | cbus_data;

#if !defined(USE_LIBFTDI)
     r = FT_SetBitMode(handle, wk, FT_BITMODE_CBUS_BITBANG);
#else
     r = ftdi_set_bitmode(handle, wk, BITMODE_CBUS); // send CBUS data
#endif
     if ((verbose>=1) || FT245R_DEBUG) {
         fprintf(stderr, "ft245r: set CBUS %x -> %d\n",dbus_data ,(int)r);
     }
  }
  if (data_changed) {
     uint8_t buf[2];
#if !defined(USE_LIBFTDI)
     FT_SetBitMode(handle, dbus_ddr, FT_BITMODE_SYNC_BITBANG);
#else
     ftdi_set_bitmode(handle, dbus_ddr, BITMODE_SYNCBB);
#endif
     buf[0] = dbus_data;
     ft245r_send (cbl, buf, 1);
     ft245r_recv (cbl, buf, 1);
  }
  return r;
}

static int setup_port(CABLE cbl, int tms, int tdi, int tck)
{
  uint8_t buf[2];

  if (tms) {
	dbus_data |= dbus_tms;
  } else {
	dbus_data &= ~dbus_tms;
  }
  if (tdi) {
	dbus_data |= dbus_tdi;
  } else {
	dbus_data &= ~dbus_tdi;
  }
  if (tck) {
	dbus_data |= dbus_tck;
  } else {
	dbus_data &= ~dbus_tck;
  }
  buf[0] = dbus_data;

  ft245r_send (cbl, buf, 1);
  ft245r_recv (cbl, buf, 1);

  return 0;
}

static int get_gpio(CABLE cbl, int bit)
{
  uint8_t buf[2];
  int ret = 0;
  int r;
  uint8_t wk;
  if (bit == CABLE_GPIO_INIT) {
	if (dbus_init) {
		buf[0] = dbus_data;
		ft245r_send (cbl, buf, 1);
		ft245r_recv (cbl, buf, 1);
		ret = (buf[0] & dbus_init) != 0;
	}
	if (cbus_init) {
		wk = (cbus_ddr << 4) | cbus_data;
#if !defined(USE_LIBFTDI)
		r = FT_SetBitMode(handle, wk, FT_BITMODE_CBUS_BITBANG);
#else
		r = ftdi_set_bitmode(handle, wk, BITMODE_CBUS);
#endif
		ret = (r & cbus_init) != 0;
	}
	ret = ret ^ init_is_neg;
  }
  return ret;
}

#if defined(USE_LIBFTDI)

#define FTDI_VID	0x0403
#define FTDI_PID	0x6001

static FT_STATUS ftdi_open_devnum(int devnum, struct ftdi_context **handlep) {
     FT_STATUS r;
     struct ftdi_context *h = NULL;
     struct ftdi_device_list *devlist = NULL, *p;
     int i;

     r = FT_ERROR;
     h = ftdi_new();
     if (!h) goto out;

     r = ftdi_usb_find_all(h, &devlist, FTDI_VID, FTDI_PID);
     if (r < 0) {
	goto out;
     }
     r = FT_ERROR;
     for (p = devlist,i=0; p; p = p->next,i++) {
	if (i == devnum) {
            r = ftdi_usb_open_dev(h, p->dev);
	    if (r == FT_OK) {
		break;
	    }
        }
     }
out:
     if (devlist) ftdi_list_free(&devlist);
     if (r == FT_OK) {
	*handlep = h;
     } else {
	if (h) ftdi_free(h);
     }
     return r;
}

static FT_STATUS ftdi_open_name(char *name, struct ftdi_context **handlep) {
     FT_STATUS r;
     struct ftdi_context *h = NULL;

     r = FT_ERROR;
     h = ftdi_new();
     if (!h) goto out;
     r = ftdi_usb_open_desc(h, FTDI_VID, FTDI_PID, name, NULL);
     if (r != FT_OK) {
         r = ftdi_usb_open_desc(h, FTDI_VID, FTDI_PID, NULL, name);
     }
out:
     if (r == FT_OK) {
	*handlep = h;
     } else {
	if (h) ftdi_free(h);
     }
     return r;
}
#endif

static int ft245r_open(CABLE cbl, char * port)
{
  FT_STATUS r = FT_OK;
  int devnum = -1;
  //uint8_t data;

  //strcpy(cbl->port, port);

  if ((strlen(port) == 3) && !strncasecmp("ft", port, 2) && '0' <= port[2] && port[2] <= '9') {
     devnum = port[2] - '0';
#if !defined(USE_LIBFTDI)
     r = FT_Open(devnum, &handle);
#else
     r = ftdi_open_devnum(devnum, &handle);
#endif
  } else {
#if !defined(USE_LIBFTDI)
     r = FT_OpenEx(port, FT_OPEN_BY_DESCRIPTION, &handle);
     if (r != FT_OK) {
        r = FT_OpenEx(port, FT_OPEN_BY_SERIAL_NUMBER, &handle);
     }
#else
     r = ftdi_open_name(port, &handle);
#endif
     if (r != FT_OK) {
        fprintf(stderr,
	    "%s: invalid portname  %s: use ft0 - ft9\n",progname,port);
        exit(1);
     }
  }
  if (r != FT_OK) {
    fprintf(stderr,
	    "%s: %s open failed \n",
	    progname, port);
    exit(1);
  }
#if !defined(USE_LIBFTDI)
  r = FT_SetLatencyTimer(handle, 2);
  r = FT_SetBitMode(handle, 0, FT_BITMODE_SYNC_BITBANG);
#else
  r = ftdi_set_bitmode(handle, 0, BITMODE_SYNCBB);
#endif
  if (r != FT_OK) {
    fprintf(stderr,
	    "%s: Synchronous BitBangMode is not supported\n",
	    progname);
     exit(1);
  }
#if FT245R_DEBUG
  fprintf(stderr, "%s: BitBang OK \n", progname);
  fflush(stderr);
#endif
  dbus_ddr = 0;
  dbus_data = 0;
  cbus_data = 0;
  setup_pin_conf(cbl, "tms", &dbus_tms, NULL, NULL);
  setup_pin_conf(cbl, "tck", &dbus_tck, NULL, NULL);
  setup_pin_conf(cbl, "tdi", &dbus_tdi, NULL, NULL);
  setup_pin_conf(cbl, "tdo", &dbus_tdo, NULL, NULL);
  setup_pin_conf(cbl, "prog", &dbus_prog, &cbus_prog, &prog_is_neg);
  setup_pin_conf(cbl, "m1", &dbus_m1, &cbus_m1, &m1_is_neg);
  setup_pin_conf(cbl, "m2", &dbus_m2, &cbus_m2, &m2_is_neg);
  setup_pin_conf(cbl, "init", &dbus_init, &cbus_init, &init_is_neg);

#if FT245R_DEBUG
  fflush(stderr);
#endif
  dbus_ddr = dbus_tck | dbus_tdi | dbus_tms;
  // set Synchronous BitBang
#if !defined(USE_LIBFTDI)
     FT_SetBitMode(handle, dbus_ddr, FT_BITMODE_SYNC_BITBANG);
#else
     ftdi_set_bitmode(handle, dbus_ddr, BITMODE_SYNCBB);
#endif
  /*
   * drain any extraneous input
   */
  drain (cbl, 0);
#if FT245R_DEBUG
  fprintf(stderr, "%s: drain OK \n", progname);
  fflush(stderr);
#endif

  is_hispeed = 0;
  valid_rates = valid_rates_r;
  valid_rates_num = sizeof(valid_rates_r)/sizeof(valid_rates_r[0]);
#if defined(USE_LIBFTDI)
#else
{
  FT_DEVICE ft_type;
  DWORD ft_id;
  DWORD trans_size = 16384 * 4;
  char sn[16];
  char desc[64];
  r = FT_GetDeviceInfo(handle, &ft_type, &ft_id, sn, desc, NULL);
  if ( (r == FT_OK) && (ft_type >= FT_DEVICE_2232H) ) {
     is_hispeed = 1;
     valid_rates = valid_rates_h;
     valid_rates_num = sizeof(valid_rates_h)/sizeof(valid_rates_h[0]);
     r = FT_SetUSBParameters(handle, trans_size, trans_size);
  }
  fflush(stderr);
}
#endif
  set_bitclock(cbl, cbl->bitclock);
  return 0;
}

static void ft245r_close(CABLE cbl, int do_reset)
{
  if (do_reset) {
#if !defined(USE_LIBFTDI)
     FT_SetBitMode(handle, 0, FT_BITMODE_RESET);
#else
     ftdi_set_bitmode(handle, 0, BITMODE_RESET);
#endif
  }
#if !defined(USE_LIBFTDI)
  FT_SetLatencyTimer(handle, 16); // default
  FT_Close(handle);  
#else
  ftdi_usb_close(handle);
  ftdi_free(handle);
  handle = NULL;
#endif
}

static struct request {
	int bytes;
	uint8_t *buf;
	int n_bits;
	struct request *next;
} *req_head,*req_tail,*req_pool;

static void put_request(int bytes, uint8_t *buf, int n_bits)
{
  struct request *p;
  if (req_pool) {
     p = req_pool;
     req_pool = p->next;
  } else {
     p = malloc(sizeof(struct request));
     if (!p) {
       fprintf(stderr, "can't alloc memory\n");
       exit(1);
     }
  }
  memset(p, 0, sizeof(struct request));
  p->bytes = bytes;
  p->buf = buf;
  p->n_bits = n_bits;
  if (req_tail) {
     req_tail->next = p;
     req_tail = p;
  } else {
     req_head = req_tail = p;
  }
}

static int do_request(CABLE cbl)
{
  struct request *p;
  int bytes;
  uint8_t mask;
  uint8_t *rcv_buf;
  int n_bits;
  int i, j, n;
  int buf_pos = 0;
  uint8_t buf[FT245R_MAX_FRAGMENT_SIZE+1+128];

  if (!req_head) return 0;
  p = req_head;
  req_head = p->next;
  if (!req_head) req_tail = req_head;

  bytes = p->bytes;
  rcv_buf = p->buf;
  n_bits = p->n_bits;

  memset(p, 0, sizeof(struct request));
  p->next = req_pool;
  req_pool = p;

  ft245r_recv(cbl, buf, bytes);
  if (!rcv_buf && n_bits) return 0;

  if (verbose>=256) fprintf(stderr, "rcv_data %d :", n_bits);
  buf_pos +=1; 
  for (i=0; i< n_bits;) {
	mask = 0x01;		// LSB first
	n = n_bits - i;
	if (n > 8) n = 8;
	*rcv_buf = 0;
	for (j=0; j< n; j++) {
		if (buf[buf_pos] & dbus_tdo) {
			if (verbose>=256) fprintf(stderr, "1");
			*rcv_buf |= mask;
		} else {
			if (verbose>=256) fprintf(stderr, "0");
		}
		buf_pos +=2; 
		mask <<= 1;
	}
	rcv_buf++;
	i += n;
  }
  if (verbose>=256) {
	fprintf(stderr, "\n");
	fflush(stderr);
  }
  return 1;
}

static inline int set_tdi(uint8_t *buf, uint8_t data, int n_bits, int tms_high)
{
    int j;
    int buf_pos = 0;
    uint8_t mask = 0x01; // LSB first

    for (j=0; j<n_bits; j++) {
	if (data & mask) {
		dbus_data |= dbus_tdi;
	} else {
		dbus_data &= ~dbus_tdi;
	}
	if (tms_high && (j == n_bits-1)) dbus_data |= dbus_tms;
	if (verbose>=256) {
		if ((dbus_data & (dbus_tms|dbus_tdi))==(dbus_tms|dbus_tdi))
			fprintf(stderr, "H");
		else if ((dbus_data & (dbus_tms|dbus_tdi))==(dbus_tms))
			fprintf(stderr, "L");
		else if ((dbus_data & (dbus_tms|dbus_tdi))==(dbus_tdi))
			fprintf(stderr, "1");
		else 
			fprintf(stderr, "0");
	}

    	buf[buf_pos] = dbus_data;
	buf_pos++;

    	buf[buf_pos] = dbus_data | dbus_tck;
	buf_pos++;

	mask <<= 1;
    }
    return buf_pos;
}

static inline int set_tms_tdi(uint8_t *buf, uint8_t data, int n_bits)
{
    int j;
    int buf_pos = 0;
    uint8_t mask = 0x01; // LSB first

    for (j=0; j<n_bits; j++) {

	if (data & mask) {	// TDI first
		dbus_data |= dbus_tdi;
	} else {
		dbus_data &= ~dbus_tdi;
	}

	mask <<= 1;

	if (data & mask) {	// TMS secound
		if (verbose>=256) fprintf(stderr, "1");
		dbus_data |= dbus_tms;
	} else {
		if (verbose>=256) fprintf(stderr, "0");
		dbus_data &= ~dbus_tms;
	}

	mask <<= 1;

    	buf[buf_pos] = dbus_data;
	buf_pos++;

    	buf[buf_pos] = dbus_data | dbus_tck;
	buf_pos++;

    }
    return buf_pos;
}

static int put_tdi_bits(CABLE cbl, uint8_t *buf, int bit_len,
				    int flags, uint8_t *rcv_buf) 
{
  uint32_t    i,j,n;
  int len=0,tms_high;
  int buf_pos = 0;
  uint8_t bb_buf[FT245R_MAX_FRAGMENT_SIZE+1+128];

  dbus_data &= ~dbus_tms; // set tms L 
  tms_high = 0;
  for (i=0; i<bit_len; ) {
     buf_pos = 0;
     if (verbose>=256) fprintf(stderr, "put_tdi_bits %d:%d ", bit_len, i);
     for (j=0; j< FT245R_FRAGMENT_SIZE/8/FT245R_CYCLES -1; j++) {
	len = bit_len - i;
	if (len > 8) len = 8;
	else {
		tms_high = flags & CABLE_TMS_HIGH;
	}
        buf_pos += set_tdi(bb_buf+buf_pos, *buf, len, tms_high); 
	buf++;
	i += len;
	if (i >= bit_len) break;
     }
     if (verbose>=256) {
	fprintf(stderr, "\n");
	fflush(stderr);
     }
     bb_buf[buf_pos++] = dbus_data; // sck down anyway
     ft245r_send(cbl, bb_buf, buf_pos);
     n = j*8;
     if (i >= bit_len) n = j*8 + len;
     put_request(buf_pos, rcv_buf, (flags & CABLE_RECIEVE)?n:0);
     rcv_buf += j;
     req_count++;
     if (req_count > REQ_OUTSTANDINGS)
        do_request(cbl);
  }
  while (do_request(cbl))
	  ;
  return bit_len;
}

static int put_tms_tdi_bits(CABLE cbl, uint8_t *buf, int bit_len,
				    int flags, uint8_t *rcv_buf) 
{
  uint32_t    i,j,n;
  int len=0;
  int buf_pos = 0;
  uint8_t bb_buf[FT245R_MAX_FRAGMENT_SIZE+1+128];

  for (i=0; i<bit_len; ) {
     buf_pos = 0;
     if (verbose>=256) fprintf(stderr, "put_tdi_tms_bits %d:%d ", bit_len, i);
     for (j=0; j< FT245R_FRAGMENT_SIZE/8/FT245R_CYCLES - 1; j++) {
	len = bit_len - i;
	if (len > 4) len = 4;
        buf_pos += set_tms_tdi(bb_buf+buf_pos, *buf, len); 
	buf++;
	i += len;
	if (i >= bit_len) break;
     }
     if (verbose>=256) {
	fprintf(stderr, "\n");
	fflush(stderr);
     }
     bb_buf[buf_pos++] = dbus_data; // sck down anyway
     ft245r_send(cbl, bb_buf, buf_pos);
     n = j*4;
     if (i >= bit_len) n = j*4 + len;
     put_request(buf_pos, rcv_buf, (flags & CABLE_RECIEVE)?n:0);
     rcv_buf += j;
     req_count++;
     if (req_count > REQ_OUTSTANDINGS)
        do_request(cbl);
  }
  while (do_request(cbl))
	  ;
  return bit_len;
}

static int get_tdo_bits(CABLE cbl, int bit_len, int tdi, uint8_t *rcv_buf)
{
  uint32_t    i,j,n;
  int len=0;
  int buf_pos = 0;
  uint8_t bb_buf[FT245R_FRAGMENT_SIZE+1+128];

  for (i=0; i<bit_len; ) {
     buf_pos = 0;
     for (j=0; j< FT245R_FRAGMENT_SIZE/8/FT245R_CYCLES -1; j++) {
	len = bit_len - i / 8;
	if (len > 8) len = 8;
        buf_pos += set_tdi(bb_buf+buf_pos, 0xff, len, 0); 
	i += len;
	if (i >= bit_len) break;
     }
     bb_buf[buf_pos++] = dbus_data; // sck down anyway
     ft245r_send(cbl, bb_buf, buf_pos);
     n = j*8;
     if (i >= bit_len) n = j*8 + len;
     put_request(buf_pos, rcv_buf, n);
     rcv_buf += j;
     req_count++;
     if (req_count > REQ_OUTSTANDINGS)
        do_request(cbl);
  }
  while (do_request(cbl))
	  ;
  return bit_len;
}


void ft245r_init(CABLE cbl)
{
  /*
   * functions
   */
  cbl->open               = ft245r_open;
  cbl->close              = ft245r_close;

  cbl->delay              = ft245r_delay;
  cbl->setup_port         = setup_port;
  cbl->put_tdi_bits       = put_tdi_bits;
  cbl->put_tms_tdi_bits   = put_tms_tdi_bits;
  cbl->get_tdo_bits       = get_tdo_bits;

  cbl->setup_gpio         = setup_gpio;
  cbl->get_gpio           = get_gpio;

  cbl->set_bitclock       = set_bitclock;
//  cbl->parseexitspecs = parseexitspecs;

}
#else
void ft245r_init(CABLE cbl)
{
}

#endif
