#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#include "util.h"
#include "cable.h"
#include "jtag.h"
#include "target.h"

/* TPI instructions */
#define TPI_SLD_POSTI 0x24 // 
#define TPI_SST_POSTI 0x64 // 
#define TPI_SETPR_LO  0x68 // 
#define TPI_SETPR_HI  0x69 // 
#define ISP_TPI_SIN   0x10 // 
#define TPI_SOUT      0x90 // 

#define NVMCSR  0x32
#define NVMCSR_BUSY 0x80

#define NVMCMD 0x33
#define NVMCMD_NOP   0x00
#define NVMCMD_WRITE 0x1d

/* TPI operations */
static uint16_t tpi_addr;

static int tpiReadCS(CABLE cbl, unsigned int address) {
	uint8_t inst;
	int s,r;
	inst = address & 0xf;
	if (address & 0x8000) { /* read CS */
		inst |= 0x80; // TPI_SLDCS
	} else { /* read IO */
		inst |= (address << 1) & 0x60;	
		inst |= 0x10; // TPI_SIN
	}
	s = rev[inst];

	jtag_SDR32(cbl, r, 16, 3, &r);
	return rev[(r >> 8) & 0xff];
}

static int tpiWriteCS(CABLE cbl, unsigned int address, uint8_t data) {
	uint8_t inst;
	int s;
	inst = address & 0xf;
	if (address & 0x8000) { /* write CS */
		inst |= 0xc0; // TPI_SSTCS
	} else {
		inst |= (address << 1) & 0x60;
		inst |= 0x90; // TPI_SOUT
	}
	s = rev[inst] | ((int)rev[data] << 8);
	jtag_SDR32(cbl, s, 16, 2, NULL);
	return 0;
}

static int tpiSetPR(CABLE cbl, uint16_t addr) {
	uint8_t inst;
	int s;
	inst = 0x68; // TPI_SETPR_LO

	s = rev[inst] | ((int)rev[addr & 0xff] << 8);
	inst = 0x69; // TPI_SETPR_HI
	s |= ((int)rev[inst] << 16) | ((int)rev[(addr >> 8) & 0xff] << 24);
	jtag_SDR32(cbl, s, 32, 2, NULL);
	return 0;
}

static int tpiReadDS(CABLE cbl, uint16_t address) {
	uint8_t inst;
	int s,r;
	if (address != tpi_addr) {
		tpiSetPR(cbl, address);
		tpi_addr = address;
	}
	inst = 0x24; // TPI_SLD_POSTI
	s = rev[inst];
	jtag_SDR32(cbl, r, 16, 3, &r);
	tpi_addr ++;
	return rev[(r >> 8) & 0xff];
}

static int tpiWriteDS(CABLE cbl, uint16_t address, uint8_t data) {
	uint8_t inst;
	int s;
	if (address != tpi_addr) {
		tpiSetPR(cbl, address);
		tpi_addr = address;
	}
	inst = 0x64; // TPI_SST_POSTI
	s = rev[inst] | ((int)rev[data] << 8);
	jtag_SDR32(cbl, s, 16, 2, NULL);
	tpi_addr ++;
	return 0;
}

int isp_program(TARGET tgt, CABLE cbl, uint8_t *buf, int len) {
	int i,r;
	int addr;
	jtag_reset(cbl);
	(tgt->sel_chan)(tgt, cbl, 2);

	r = tpiReadCS(cbl, NVMCSR);
	tpiWriteCS(cbl, NVMCMD, NVMCMD_WRITE);
	r = tpiReadCS(cbl, NVMCSR);
	if (r & NVMCSR_BUSY) {
		return -1;
	}
	tpi_addr = 0;
	addr = 0x4000;
	for (i=0; i<len; i++) {
		tpiWriteDS(cbl, addr, buf[i]);
		addr++;
	}
	tpiWriteCS(cbl, NVMCMD, NVMCMD_NOP);
	return 0;
}

int isp_verify(TARGET tgt, CABLE cbl, uint8_t *buf, int len) {
	int i,r;
	int addr;
	jtag_reset(cbl);
	(tgt->sel_chan)(tgt, cbl, 2);

	tpi_addr = 0;
	addr = 0x4000;
	for (i=0; i<len; i++) {
		r = tpiReadDS(cbl, addr);
		if (r != buf[i]) return -i;
		addr++;
	}
	return i;
}

int isp_read(TARGET tgt, CABLE cbl, uint8_t *buf, int len) {
	int i;
	int addr;
	jtag_reset(cbl);
	(tgt->sel_chan)(tgt, cbl, 2);

	tpi_addr = 0;
	addr = 0x4000;
	for (i=0; i<len; i++) {
		buf[i] = tpiReadDS(cbl, addr);
		addr++;
	}
	return i;
}
