#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#include "util.h"
#include "cable.h"
#include "jtag.h"

extern int verbose;

static int jtag_state = JTAG_STATE_UNKOWN;

static char * state_str(int st) {
   char *str;
   switch (st) {
	case JTAG_STATE_IDLE:
		str = "JTAG_STATE_IDLE";
		break;
	case JTAG_STATE_SIR:
		str = "JTAG_STATE_SIR";
		break;
	case JTAG_STATE_SDR:
		str = "JTAG_STATE_SDR";
		break;
	case JTAG_STATE_EXIT1_IR:
		str = "JTAG_STATE_EXIT1_IR";
		break;
	case JTAG_STATE_EXIT1_DR:
		str = "JTAG_STATE_EXIT1_DR";
		break;
	case JTAG_STATE_UNKOWN:
	default:
		str = "JTAG_STATE_UNKOWN";
   }
   return str;
}

void jtag_go(CABLE cbl,int state) {
	int n = 0;
	uint8_t buf[4];
	uint32_t code = 0;
	
	if (state == JTAG_STATE_IDLE) {
		if (jtag_state == JTAG_STATE_IDLE) return;
		if ( (jtag_state == JTAG_STATE_SIR)
		    || (jtag_state == JTAG_STATE_SDR) ) {
		    n = 3;
		    code = 0x0A;		// TMS = 110
		} else if ( (jtag_state == JTAG_STATE_EXIT1_IR)
		          || (jtag_state == JTAG_STATE_EXIT1_DR) ) {
		    n = 4;
		    code = 0x28;		// TMS = 0110
		} else {
		    n = 6;
		    code = 0x02AA;		// TMS = 111110
		}
	} else if (state == JTAG_STATE_SIR) { // (really Capture-IR)
		if (jtag_state == JTAG_STATE_SIR) return;
		if (jtag_state == JTAG_STATE_IDLE) {
		    n = 4;
		    code = 0x0A;		// TMS = 1100
		} else if ( (jtag_state == JTAG_STATE_EXIT1_IR)
		          || (jtag_state == JTAG_STATE_EXIT1_DR) ) {
		    n = 8;
		    code = 0x0A28;		// TMS = 01101100
		} else if (jtag_state == JTAG_STATE_SDR) {
		    n = 7;
		    code = 0x028A;		// TMS = 1101100
		} else {
		    n = 10;
		    code = 0x00A2AA;		// TMS = 1111101100
		}
	} else if (state == JTAG_STATE_SDR) {
		if (jtag_state == JTAG_STATE_SDR) return;
		if (jtag_state == JTAG_STATE_IDLE) {
		    n = 3;
		    code = 0x02;		// TMS = 100
		} else if ( (jtag_state == JTAG_STATE_EXIT1_IR)
		          || (jtag_state == JTAG_STATE_EXIT1_DR) ) {
		    n = 7;
		    code = 0x0228;		// TMS = 0110100
		} else if (jtag_state == JTAG_STATE_SIR) {
		    n = 7;
		    code = 0x020A;		// TMS = 1100100
		} else {
		    n = 10;
		    code = 0x0022AA;		// TMS = 111110100
		}
	}
	buf[0] = code & 0xff;
	buf[1] = (code >> 8) & 0xff;
	buf[2] = (code >> 16) & 0xff;
	buf[3] = (code >> 24) & 0xff;
	if (verbose>=256) {
		fprintf(stderr, "jtag_go %s - %s : %d %06x\n"
				, state_str(jtag_state), state_str(state)
				, n , code);
	}
	(cbl->put_tms_tdi_bits)(cbl, buf, n, 0 , NULL);
	jtag_state = state;
}

void jtag_init(CABLE cbl) {
	jtag_state = JTAG_STATE_UNKOWN;
	(cbl->setup_port)(cbl, 1, 0, 0); // TMS = 1, TDI = 0, TCK = 0;
	jtag_go(cbl, JTAG_STATE_IDLE);
}

void jtag_reset(CABLE cbl) {
	jtag_state = JTAG_STATE_UNKOWN;
	jtag_go(cbl, JTAG_STATE_IDLE);
}

#define IDLE_BUF_SIZE	8192
void jtag_IDLE(CABLE cbl, int ms, int count) {
	static uint8_t zero[IDLE_BUF_SIZE];
	int subcnt,tmp;
	int save_verb = verbose;
	tmp = cbl->bitclock * ms / 1000.0;
	if (tmp > count) count = tmp;

	jtag_go(cbl, JTAG_STATE_IDLE);
	if (verbose >=256) {
		verbose = 1;
	}
	while (count) {
		subcnt = count;
		if (subcnt > IDLE_BUF_SIZE * 8) subcnt = IDLE_BUF_SIZE * 8;
		count -= subcnt;
		(cbl->put_tdi_bits)(cbl, zero, subcnt, 0 , NULL);
	}
	verbose = save_verb;
}

void jtag_SIR(CABLE cbl, uint8_t *buf, int bit_len, int flags, uint8_t *rcv_buf) {
	int c_flags = 0;
	jtag_go(cbl, JTAG_STATE_SIR);
	if (flags & 1) {
		memset(rcv_buf, 0, (bit_len + 7)/8);
		c_flags |= CABLE_RECIEVE;
	}
	if (!(flags & 2)) {
		c_flags |= CABLE_TMS_HIGH;
		jtag_state = JTAG_STATE_EXIT1_IR;
	}
	(cbl->put_tdi_bits)(cbl, buf, bit_len, c_flags , rcv_buf);
}

void jtag_SDR(CABLE cbl, uint8_t *buf, int bit_len, int flags, uint8_t *rcv_buf) {
	int c_flags = 0;
	jtag_go(cbl, JTAG_STATE_SDR);
	if (flags & 1) {
		c_flags |= CABLE_RECIEVE;
	}
	if (!(flags & 2)) {
		c_flags |= CABLE_TMS_HIGH;
		jtag_state = JTAG_STATE_EXIT1_DR;
	}
	(cbl->put_tdi_bits)(cbl, buf, bit_len, c_flags , rcv_buf);
}

void jtag_SIR32(CABLE cbl, uint32_t code, int bit_len, int flags, uint32_t *rcv_data) {
	uint8_t buf[4];
	uint8_t rcv_buf[4];
	buf[0] = code;
	buf[1] = code >> 8;
	buf[2] = code >> 16;
	buf[3] = code >> 24;

	memset(rcv_buf, 0 , 4);
	jtag_SIR(cbl, buf, bit_len, flags, rcv_buf);
	if (flags & 1) {
		*rcv_data = rcv_buf[0] | ((uint32_t)rcv_buf[1]<<8) 
			| ((uint32_t)rcv_buf[2]<<16)
			| ((uint32_t)rcv_buf[3]<<24);
	}
}

void jtag_SDR32(CABLE cbl, uint32_t code, int bit_len, int flags, uint32_t *rcv_data) {
	uint8_t buf[4];
	uint8_t rcv_buf[4];
	buf[0] = code;
	buf[1] = code >> 8;
	buf[2] = code >> 16;
	buf[3] = code >> 24;

	memset(rcv_buf, 0 , 5);
	jtag_SDR(cbl, buf, bit_len, flags, rcv_buf);
	if (flags & 1) {
		*rcv_data = rcv_buf[0] | ((uint32_t)rcv_buf[1]<<8) 
			| ((uint32_t)rcv_buf[2]<<16)
			| ((uint32_t)rcv_buf[3]<<24);
	}
}

void jtag_SDR32R(CABLE cbl, uint32_t code, int bit_len, int flags, uint32_t *rcv_data) {
	int i;
	uint8_t buf[4];
	uint8_t rcv_buf[4];
	uint32_t r,r2;

	code <<= (32 - bit_len);
	buf[0] = code >> 24;
	buf[1] = code >> 16;
	buf[2] = code >> 8;
	buf[3] = code;
	for (i=0; i<4; i++) buf[i] = rev[buf[i]];

	memset(rcv_buf, 0 , 4);
	jtag_SDR(cbl, buf, bit_len, flags, rcv_buf);
//printf("  code %02x ", code);
	if (flags & 1) {
		for (i=0; i<4; i++) rcv_buf[i] = rev[rcv_buf[i]];
		r = rcv_buf[3] | ((uint32_t)rcv_buf[2]<<8) 
			| ((uint32_t)rcv_buf[1]<<16)
			| ((uint32_t)rcv_buf[0]<<24);
		r2 = r >> (32 - bit_len);
		*rcv_data = r2;
//printf("  r %08x *rcv %02x (sft %d)", r, r2,  (32 - bit_len));
	}
//printf("\n");
}
