#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>
#include <malloc.h>
#include <fcntl.h>

#ifdef __linux__
#include <termios.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#define O_BINARY	0

#elif defined(WIN32NATIVE)

#include <windows.h>
#include <io.h>
#ifdef HAVE_TERMIOS
#include <termios.h>
#else
#include <wincon.h>
#endif

#endif

#include "util.h"

#if defined(__linux__) || defined(WIN32NATIVE)

#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
static HANDLE conin;
static DWORD conin_mode_sv;
#else
struct termios term_sv;
#endif

void term_exit() {
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	if (conin_mode_sv != 0) 
		SetConsoleMode(conin, conin_mode_sv);
#else
	tcsetattr(0, TCSANOW, &term_sv);
#endif
}

void term_enter() {
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	DWORD conin_mode = 0;
	int r;

	conin = GetStdHandle(STD_INPUT_HANDLE);
	r = GetConsoleMode(conin, &conin_mode);
	if (r) {
		conin_mode_sv = conin_mode;
		conin_mode &= ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT);
		r = SetConsoleMode(conin, conin_mode);
	} else {
		fprintf(stderr, "run on the console(command prompt)\n");
		exit(1);
	}
#else
	struct termios term;
	tcgetattr(0, &term);
	term_sv = term;
	term.c_lflag &= ~(ECHO | ICANON);
	term.c_iflag &= ~(ICRNL);
	tcsetattr(0, TCSANOW, &term);
#endif
	atexit(term_exit);

fprintf(stderr,"end lin_init\n");
fflush(stderr);
}

void term_putc(char data) {
	write(1, &data, 1);
}

void term_puts(char *s) {
	int len = strlen(s);
	write(1, s, len);
}

char term_getc(void) {
	int r;
	char ret = 0;
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	char buf[3];
	DWORD nrec = 0;

	while (!ReadConsole(conin, buf, 1, &nrec, NULL) || nrec == 0)
			;
		return buf[0];
#endif
	while ((r = read(0, &ret, 1)) != 1)
		;

	return ret;
}

#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
static int8_t term_can_getc(int8_t size) {
	INPUT_RECORD ir;
	DWORD nrec = 0;
	if (PeekConsoleInput(conin, &ir, 1, &nrec)) {
		if ((nrec != 0) && (ir.EventType & KEY_EVENT) &&
			ir.Event.KeyEvent.bKeyDown ) {
fprintf(stderr,"Console %d %08x\n",ir.Event.KeyEvent.bKeyDown, ir.Event.KeyEvent.wVirtualKeyCode);
fflush(stderr);
			return 1;
		}
	}
	return 0;
}
#else
#include <sys/time.h>
#if defined(WIN32NATIVE) && defined(HAVE_TERMIOS)
#define select _select
#endif
static int8_t term_can_getc(int8_t size) {
	int r;
	fd_set fds;
	struct timeval to;
	to.tv_usec = 50000; // 50 ms
	to.tv_sec = 0;

	FD_ZERO(&fds);
	FD_SET(0, &fds);

	r = select(1, &fds, NULL, NULL, &to);
	if (r == 1) {
		return 1;
	}
	return 0;
}
#endif
#endif

#include "cable.h"
#include "jtag.h"
#include "target.h"

void term_mode(TARGET tgt, CABLE cbl) {
	int c_prev2 = -1;
	int c_prev = -1;
	int null_count = 0;
	int c,r;

	jtag_reset(cbl);
	(tgt->sel_chan)(tgt, cbl, 1);
	term_enter();
	term_puts("\r\nEnter SPI Terminal Mode\r\n");
	for (;;) {
		if (term_can_getc(1)) {
			c = rev[term_getc() & 0x7f];
		} else {
			c = 0xff;
		}
		jtag_SDR32(cbl, c, 8, 3, &r);
		if (!(r & 1)) {
			term_putc(rev[r & 0xff]);
		}
		if ((c_prev2 == '+') && (c_prev == '+') && (c == '+')) {
			break;
		}
		if ((c_prev == '~') && (c == '.')) {
			break;
		}
		c_prev2 = c_prev;
		c_prev = c;
		if ((c == 0xff) && (r & 1)) {
			null_count ++;
			if (null_count > 20) {
				jtag_SDR32(cbl, 0xff, 8, 1, &r);
				jtag_IDLE(cbl, 20, 5);
				if (!(r & 1)) {
					term_putc(rev[r & 0xff]);
				}
				usleep(200000); // 200ms
				c_prev2 = -1;
				c_prev = -1;
				null_count = 0;
			}
		} else {
			null_count = 0;
		}
	}
	term_puts("\r\nExiting SPI Terminal Mode\r\n");
	term_exit();
}
